use Test::More tests => 1;

BEGIN {
use_ok( 'Maypole::Plugin::QuickTable' );
}

diag( "Testing Maypole::Plugin::QuickTable $Maypole::Plugin::QuickTable::VERSION, Perl 5.008006, /usr/local/bin/perl" );
